use crate::TallyState;
use futures::stream;
use futures::sync::mpsc::Receiver;
use futures::sync::oneshot;
use std::sync::{Arc, Mutex};
use std::thread;
use typed_html::dom::DOMTree;
use typed_html::html;
use warp::filters::ws::Message;
use warp::{Filter, Future, Stream};

fn build_template() -> DOMTree<String> {
    html!(
        <html>
            <head>
                <title>"OpenTally Control"</title>
                <script src="/script.js" type="text/javascript"></script>
            </head>
            <body>
                <h1>"OpenTally Control"</h1>
            </body>
        </html>
    )
}

pub struct WebServer {
    shutdown_chan: oneshot::Sender<()>,
    shutdown_mutex: Arc<Mutex<()>>,
}

impl WebServer {
    pub fn run(addr: std::net::SocketAddr, tally_rx: Receiver<TallyState>) -> WebServer {
        // todo: not have the damn js inline in the func
        const JS: &'static str = r#"
var sock = new WebSocket("ws://" + window.location.host + "/ws");
sock.onmessage = function (e) {
    console.log(e);
}
sock.onopen = function (e) {
    sock.send("Hello!");
}"#;
        let (shutdown_tx, shutdown_rx) = oneshot::channel::<()>();

        let shutdown_mutex = Arc::new(Mutex::new(()));
        let thread_shutdown_mutex = Arc::clone(&shutdown_mutex);
        thread::spawn(move || {
            // N.B! Despite the fact we never use the value inside the mutex (it's () after all..) we need to bind
            // it to a variable so that it lives until the end of the thread, else it would unlock immediately.
            let _lock = thread_shutdown_mutex.lock().unwrap();

            let websocket = warp::path("ws").and(warp::ws2()).map(|ws: warp::ws::Ws2| {
                ws.on_upgrade(|websocket| {
                    let (tx, _) = websocket.split();
                    //stream::once::<Message, _>(Ok(Message::text("Hello")))
                    tally_rx
                        .map(|state| Message::text(format!("{}", state)))
                        .forward(tx)
                        .map(|_| ())
                        .map_err(|e| {
                            log::error!("Websocket error: {}", e);
                        })
                })
            });
            let script = warp::path("script.js").map(|| JS);
            let home = warp::path::end().map(|| warp::reply::html(build_template().to_string()));
            let routes = warp::get2().and(websocket.or(script).or(home));

            let (_addr, server) =
                warp::serve(routes).bind_with_graceful_shutdown(addr, shutdown_rx);
            tokio::run(server);
        });
        WebServer {
            shutdown_chan: shutdown_tx,
            shutdown_mutex,
        }
    }

    pub fn shutdown(self) {
        log::debug!("Web server shutting down...");
        self.shutdown_chan.send(()).unwrap();
        let _lock = self.shutdown_mutex.lock().unwrap();
        log::info!("Web server shutdown");
    }
}
